package org.jastadd.ros2rag.tests;

import com.google.protobuf.InvalidProtocolBufferException;
import config.Robotconfig.RobotConfig;
import example.ast.*;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

import static org.junit.Assert.*;

/**
 * Test case "example".
 *
 * @author rschoene - Initial contribution
 */
public class ExampleTest extends AbstractMqttTest {

  private static final String TOPIC_CONFIG = "robot/config";
  private static final String TOPIC_JOINT1 = "robot/arm/joint1";

  private Model model;
  private RobotArm robotArm;
  private Joint joint1;
  private MqttUpdater receiver;

  @AfterEach
  public void closeConnections() {
    if (receiver != null) {
      receiver.close();
    }
    if (model != null) {
      model.MqttCloseConnections();
    }
  }

  @Test
  public void buildModel() {
    createModel();
  }

  @Test
  public void communicate() throws IOException, InterruptedException {
    createModel();
    List<RobotConfig> receivedConfigs = new ArrayList<>();

    createListener(receivedConfigs);

    model.MqttSetHost(TestUtils.getMqttHost());
    assertTrue(model.MqttWaitUntilReady(2, TimeUnit.SECONDS));
    // joint is currently within the safety zone, so speed should be low
    robotArm.connectAppropriateSpeed(TOPIC_CONFIG, true);
    joint1.connectCurrentPosition(TOPIC_JOINT1);

    // now change the position of the joint out of the safety zone
    joint1.setCurrentPosition(makePosition(2, 2, 2));

    // and wait for MQTT to send/receive messages
    TimeUnit.SECONDS.sleep(2);

    // there should be two configs received by now (the initial one, and the updated)
    assertEquals(2, receivedConfigs.size());
    RobotConfig actualInitialConfig = receivedConfigs.get(0);
    assertEquals(robotArm.speedLow(), actualInitialConfig.getSpeed(), TestUtils.DELTA);

    RobotConfig actualUpdatedConfig = receivedConfigs.get(1);
    assertEquals(robotArm.speedHigh(), actualUpdatedConfig.getSpeed(), TestUtils.DELTA);
  }

  private void createListener(List<RobotConfig> receivedConfigs) {
    receiver = new MqttUpdater("receiver");
    try {
      receiver.setHost(TestUtils.getMqttHost(), TestUtils.getMqttDefaultPort());
    } catch (IOException e) {
      fail("Could not set host: " + e.getMessage());
    }
    assertTrue(receiver.waitUntilReady(2, TimeUnit.SECONDS));
    receiver.newConnection(TOPIC_CONFIG, bytes -> {
      try {
        RobotConfig config = RobotConfig.parseFrom(bytes);
        receivedConfigs.add(config);
      } catch (InvalidProtocolBufferException e) {
        fail("Received bad config: " + e.getMessage());
      }
    });
  }

  private void createModel() {
    model = new Model();

    ZoneModel zoneModel = new ZoneModel();
    zoneModel.setSize(makePosition(1, 1, 1));

    IntPosition myPosition = makePosition(0, 0, 0);
    Coordinate myCoordinate = new Coordinate(myPosition);
    Coordinate leftPosition = new Coordinate(makePosition(-1, 0, 0));
    Coordinate rightPosition = new Coordinate(makePosition(1, 0, 0));

    Zone safetyZone = new Zone();
    safetyZone.addCoordinate(myCoordinate);
    safetyZone.addCoordinate(leftPosition);
    safetyZone.addCoordinate(rightPosition);
    zoneModel.addSafetyZone(safetyZone);
    model.setZoneModel(zoneModel);

    robotArm = new RobotArm();
    robotArm.setAttributeTestSource(1);  // set initial value, no trigger

    joint1 = new Joint();
    joint1.setName("joint1");
    joint1.setCurrentPosition(myPosition);

    EndEffector endEffector = new EndEffector();
    endEffector.setName("gripper");
    endEffector.setCurrentPosition(makePosition(2, 2, 3));

    robotArm.addJoint(joint1);
    robotArm.setEndEffector(endEffector);
    model.setRobotArm(robotArm);

    // add dependencies
    robotArm.addDependency1(joint1);
    robotArm.addDependency1(endEffector);
    robotArm.addDependency2(robotArm);
  }

  private static IntPosition makePosition(int x, int y, int z) {
    return IntPosition.of(x, y, z);
  }

}
